/*
 * Copyright (c) 2013, Pushkar Singh <psingh@sangoma.com>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 * * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * * Neither the name of the original author; nor the names of any contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 * 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors: Moises Silva  <moy@sangoma.com>
 * 		 Pushkar Singh <psingh@sangoma.com>
 *
 */

/********************************************************************************/
/*                                                                              */
/*                           libraries inclusion                                */
/*                                                                              */
/********************************************************************************/
#define _GNU_SOURCE
#ifndef WIN32
#include <poll.h>
#endif

#include "private/ftdm_core.h"
#include <stdarg.h>
#include <ctype.h>
#ifdef WIN32
#include <io.h>
#endif
#include "sng_decoder.h"
#include "ftmod_sangoma_isdn_tap.h"

/********************************************************************************/
/*                                                                              */
/*                               global declaration                             */
/*                                                                              */
/********************************************************************************/
static int g_decoder_lib_init;
static sng_decoder_event_interface_t    sng_event;
static ftdm_io_interface_t ftdm_sngisdn_tap_interface;
static ftdm_status_t ftdm_sangoma_isdn_tap_start(ftdm_span_t *span);
void sangoma_isdn_tap_set_debug(struct sngisdn_t *pri, int debug);
static void handle_isdn_tapping_log(uint8_t level, char *fmt,...);
int ftdm_sangoma_isdn_get_switch_type(const char *val);
static void sangoma_isdn_tap_check_event(ftdm_span_t* span);

/********************************************************************************/
/*                                                                              */
/*                              implementation                                  */
/*                                                                              */
/********************************************************************************/

/**********************************************************************************
 * Fun   : FIO_IO_UNLOAD_FUNCTION(ftdm_sangoma_isdn_tap_unload)
 *
 * Desc  : Call back function in order to unload the sng_isdn_tap module
 *
 * Ret   : FTDM_SUCCESS
**********************************************************************************/
static FIO_IO_UNLOAD_FUNCTION(ftdm_sangoma_isdn_tap_unload)
{
	/* Release Decoder lib resources */
	if(g_decoder_lib_init) {
		sng_decoder_lib_deinit();
		g_decoder_lib_init = 0x00;
	}
	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : FIO_CHANNEL_GET_SIG_STATUS_FUNCTION(sangoma_isdn_tap_get_channel_sig_status)
 * 
 * Desc  : Call back function in order to get the signalling status w.r.t a 
 * 	   particular channel as requested by user
 *
 * Ret   : FTDM_SUCCESS
**********************************************************************************/
static FIO_CHANNEL_GET_SIG_STATUS_FUNCTION(sangoma_isdn_tap_get_channel_sig_status)
{
	if (ftdm_test_flag(ftdmchan, FTDM_CHANNEL_SIG_UP)) {
		*status = FTDM_SIG_STATE_UP;
	} else {
		*status = FTDM_SIG_STATE_DOWN;
	}

	*status = FTDM_SIG_STATE_UP;
	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : FIO_SPAN_GET_SIG_STATUS_FUNCTION(sangoma_isdn_tap_get_span_sig_status)
 * 
 * Desc  : Call back function in order to get the signalling status w.r.t a 
 * 	   particular span as requested by user
 *
 * Ret   : FTDM_SUCCESS
**********************************************************************************/
static FIO_SPAN_GET_SIG_STATUS_FUNCTION(sangoma_isdn_tap_get_span_sig_status)
{
	*status = FTDM_SIG_STATE_UP;
	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : FIO_CHANNEL_OUTGOING_CALL_FUNCTION(sangoma_isdn_tap_outgoing_call)
 * 
 * Desc  : Call back function for outgoing call but in case of tapping one cannot
 * 	   perform any outgoing calls
 *
 * Ret   : FTDM_FAIL
**********************************************************************************/
static FIO_CHANNEL_OUTGOING_CALL_FUNCTION(sangoma_isdn_tap_outgoing_call)
{
	ftdm_log(FTDM_LOG_ERROR, "Cannot dial on ISDN tapping line!\n");
	return FTDM_FAIL;
}

/**********************************************************************************
 * Fun   : parse_debug()
 * 
 * Desc  : Sets the dubug level as per users requirement
 *
 * Ret   : DEBUG_ALL | flags | 0
**********************************************************************************/
static int parse_debug(const char *in)
{
	int flags = 0;

	if (!in) {
		return 0;
	}

	if (!strcmp(in, "none")) {
		return 0;
	}

	if (!strcmp(in, "all")) {
		return DEBUG_ALL;
	}

	if (strstr(in, "q921_raw")) {
		flags |= DEBUG_Q921_RAW;
	}

	return flags;
}

/**********************************************************************************
 * Fun   : FIO_API_FUNCTION(ftdm_sangoma_isdn_tap_api)
 * 
 * Desc  : Call back function in order provide CLI commands w.r.t dng_isdn_tap 
 * 	   module
 *
 * Ret   : FTDM_SUCCESS
**********************************************************************************/
static FIO_API_FUNCTION(ftdm_sangoma_isdn_tap_api)
{
	char *mycmd = NULL, *argv[10] = { 0 };
	int argc = 0;
	
	if (data) {
		mycmd = ftdm_strdup(data);
		argc = ftdm_separate_string(mycmd, ' ', argv, (sizeof(argv) / sizeof(argv[0])));
	}

	if (argc > 2) {
		if (!strcasecmp(argv[0], "debug")) {
			ftdm_span_t *span = NULL;

			if (ftdm_span_find_by_name(argv[1], &span) == FTDM_SUCCESS) {
				sngisdn_tap_t *tap = span->signal_data;
				if (span->start != ftdm_sangoma_isdn_tap_start) {
					stream->write_function(stream, "%s: -ERR invalid span.\n", __FILE__);
					goto done;
				}

				/* Setting Up the debug value */
				sangoma_isdn_tap_set_debug(tap->pri, parse_debug(argv[2]));
				stream->write_function(stream, "%s: +OK debug set.\n", __FILE__);
				goto done;
			} else {
				stream->write_function(stream, "%s: -ERR invalid span.\n", __FILE__);
				goto done;
			}
		}

	}

	stream->write_function(stream, "%s: -ERR invalid command.\n", __FILE__);
	
 done:
	ftdm_safe_free(mycmd);
	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : FIO_IO_LOAD_FUNCTION(ftdm_sangoma_isdn_tap_io_init)
 * 
 * Desc  : Call back function in order to initialize set io w.r.t sng_isdn_tap
 *
 * Ret   : FTDM_SUCCESS
**********************************************************************************/
static FIO_IO_LOAD_FUNCTION(ftdm_sangoma_isdn_tap_io_init)
{
	memset(&ftdm_sngisdn_tap_interface, 0, sizeof(ftdm_sngisdn_tap_interface));

	ftdm_sngisdn_tap_interface.name = "sangoma_isdn_tap";
	ftdm_sngisdn_tap_interface.api = ftdm_sangoma_isdn_tap_api;

	*fio = &ftdm_sngisdn_tap_interface;

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : FIO_SIG_LOAD_FUNCTION(ftdm_sangoma_isdn_tap_init)
 * 
 * Desc  : Call back function in order to initialize sngisdn_tap module + stack +
 * 	   decoder library
 *
 * Ret   : FTDM_SUCCESS | FTDM_FAIL
**********************************************************************************/
static FIO_SIG_LOAD_FUNCTION(ftdm_sangoma_isdn_tap_init)
{
	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Macro : ftdm_state_map_t
 * 
 * Desc  : Mapping FTDM events
**********************************************************************************/
static ftdm_state_map_t sngisdn_tap_state_map = {
	{
		{
			ZSD_INBOUND,
			ZSM_UNACCEPTABLE,
			{FTDM_CHANNEL_STATE_DOWN, FTDM_END},
			{FTDM_CHANNEL_STATE_RING, FTDM_END}
		},
		{
			ZSD_INBOUND,
			ZSM_UNACCEPTABLE,
			{FTDM_CHANNEL_STATE_RING, FTDM_END},
			{FTDM_CHANNEL_STATE_HANGUP, FTDM_CHANNEL_STATE_TERMINATING, FTDM_CHANNEL_STATE_PROGRESS, FTDM_CHANNEL_STATE_PROGRESS_MEDIA, FTDM_CHANNEL_STATE_UP, FTDM_END}
		},
		{
			ZSD_INBOUND,
			ZSM_UNACCEPTABLE,
			{FTDM_CHANNEL_STATE_HANGUP, FTDM_END},
			{FTDM_CHANNEL_STATE_TERMINATING, FTDM_END},
		},
		{
			ZSD_INBOUND,
			ZSM_UNACCEPTABLE,
			{FTDM_CHANNEL_STATE_TERMINATING, FTDM_END},
			{FTDM_CHANNEL_STATE_DOWN, FTDM_END},
		},
		{
			ZSD_INBOUND,
			ZSM_UNACCEPTABLE,
			{FTDM_CHANNEL_STATE_PROGRESS, FTDM_END},
			{FTDM_CHANNEL_STATE_HANGUP, FTDM_CHANNEL_STATE_TERMINATING, FTDM_CHANNEL_STATE_PROGRESS_MEDIA, FTDM_CHANNEL_STATE_UP, FTDM_END},
		},
		{
			ZSD_INBOUND,
			ZSM_UNACCEPTABLE,
			{FTDM_CHANNEL_STATE_PROGRESS_MEDIA, FTDM_END},
			{FTDM_CHANNEL_STATE_HANGUP, FTDM_CHANNEL_STATE_TERMINATING, FTDM_CHANNEL_STATE_UP, FTDM_END},
		},
		{
			ZSD_INBOUND,
			ZSM_UNACCEPTABLE,
			{FTDM_CHANNEL_STATE_UP, FTDM_END},
			{FTDM_CHANNEL_STATE_HANGUP, FTDM_CHANNEL_STATE_TERMINATING, FTDM_END},
		},
		
	}
};

/**********************************************************************************
 * Fun   : ftdm_status_t state_advance
 * 
 * Desc  : Change FTDM state depending upon Q931 message received
 *
 * Ret   : FTDM_SUCCESS
**********************************************************************************/
#define SNG_ISDN_TAP_GET_INTERFACE(iface) iface == ISDN_TAP_IFACE_CPE ? "CPE" :  \
                                    iface == ISDN_TAP_IFACE_NET ? "NET" : "UNKNOWN"
static ftdm_status_t state_advance(ftdm_channel_t *ftdmchan)
{
	ftdm_status_t status;
	ftdm_sigmsg_t sig;
#if 0
	sngisdn_tap_t *tap = ftdmchan->span->signal_data;
        sngisdn_tap_t *peer_tap = tap->peerspan->signal_data;
#endif
	ftdm_log_chan(ftdmchan, FTDM_LOG_DEBUG, "processing state %s\n", ftdm_channel_state2str(ftdmchan->state));

	memset(&sig, 0, sizeof(sig));
	sig.chan_id = ftdmchan->chan_id;
	sig.span_id = ftdmchan->span_id;
	sig.channel = ftdmchan;

	ftdm_channel_complete_state(ftdmchan);

	switch (ftdmchan->state) {
	case FTDM_CHANNEL_STATE_DOWN:
		{			
			ftdm_channel_t *fchan = ftdmchan;

			/* Destroy the peer data first */
			if (fchan->call_data) {
				ftdm_channel_t *peerchan = fchan->call_data;
				ftdm_channel_t *pchan = peerchan;

				ftdm_channel_lock(peerchan);

				pchan->call_data = NULL;
				pchan->pflags = 0;
				ftdm_channel_close(&pchan);

				ftdm_channel_unlock(peerchan);
			} else {
				ftdm_log_chan_msg(ftdmchan, FTDM_LOG_CRIT, "No call data?\n");
			}

			ftdmchan->call_data = NULL;
			ftdmchan->pflags = 0;
			ftdm_channel_close(&fchan);
		}
		break;

	case FTDM_CHANNEL_STATE_PROGRESS:
	case FTDM_CHANNEL_STATE_PROGRESS_MEDIA:
	case FTDM_CHANNEL_STATE_HANGUP:
		break;

	case FTDM_CHANNEL_STATE_UP:
		{
			/* Start Tapping Call as call is in connected state */
				sig.event_id = FTDM_SIGEVENT_UP;
				if ((status = ftdm_span_send_signal(ftdmchan->span, &sig) != FTDM_SUCCESS)) {
					ftdm_set_state_locked(ftdmchan, FTDM_CHANNEL_STATE_HANGUP);
				}
		}
		break;

	case FTDM_CHANNEL_STATE_RING:
		{
			/* We will start Tapping calls only when Connect is recieved */
#if 0
			sig.event_id = FTDM_SIGEVENT_START;
			/* The ring interface (where the setup was received) is the peer, since we RING the channel
			 * where PROCEED/PROGRESS is received */
			ftdm_sigmsg_add_var(&sig, "sngisdn_tap_ring_interface", SNG_ISDN_TAP_GET_INTERFACE(peer_tap->iface));
			if ((status = ftdm_span_send_signal(ftdmchan->span, &sig) != FTDM_SUCCESS)) {
				ftdm_set_state_locked(ftdmchan, FTDM_CHANNEL_STATE_HANGUP);
			}
#endif
		}
		break;

	case FTDM_CHANNEL_STATE_TERMINATING:
		{
			if (ftdmchan->last_state != FTDM_CHANNEL_STATE_HANGUP) {
				sig.event_id = FTDM_SIGEVENT_STOP;
				status = ftdm_span_send_signal(ftdmchan->span, &sig);
			}
			ftdm_set_state_locked(ftdmchan, FTDM_CHANNEL_STATE_DOWN);
		}
		break;

	default:
		{
			ftdm_log_chan(ftdmchan, FTDM_LOG_DEBUG, "ignoring state change from %s to %s\n", ftdm_channel_state2str(ftdmchan->last_state), ftdm_channel_state2str(ftdmchan->state));
		}
		break;
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : ftmod_isdn_process_event()
 *
 * Desc  : Handle wanpipe events does not change state from Down to Up once ALARM
 * 	   Clear is recieved as We only change channel state to UP once call is
 * 	   being answered on that particular channel and Tapping starts
 *
 * Ret   : FTDM_SUCCESS
 **********************************************************************************/
static __inline__ ftdm_status_t ftmod_isdn_process_event(ftdm_span_t *span, ftdm_event_t *event)
{
	ftdm_alarm_flag_t alarmbits;
	ftdm_sigmsg_t sig;

	memset(&sig, 0, sizeof(sig));
	sig.span_id = ftdm_channel_get_span_id(event->channel);
	sig.chan_id = ftdm_channel_get_id(event->channel);
	sig.channel = event->channel;

	ftdm_log_chan(event->channel, FTDM_LOG_DEBUG, "EVENT:%d Recieved on Span:%d Channel:%d\n",
			event->enum_id,	ftdm_channel_get_span_id(event->channel),
			ftdm_channel_get_id(event->channel));

	switch (event->enum_id) {
		case FTDM_OOB_ALARM_TRAP:
			{
				sig.event_id = FTDM_OOB_ALARM_TRAP;
				if (ftdm_channel_get_state(event->channel) != FTDM_CHANNEL_STATE_DOWN) {
					ftdm_set_state_locked(event->channel, FTDM_CHANNEL_STATE_RESTART);
				}
				ftdm_set_flag(event->channel, FTDM_CHANNEL_SUSPENDED);
				ftdm_channel_get_alarms(event->channel, &alarmbits);
				ftdm_span_send_signal(span, &sig);

				ftdm_log(FTDM_LOG_WARNING, "channel %d:%d (%d:%d) has alarms [%s]\n",
						ftdm_channel_get_span_id(event->channel),
						ftdm_channel_get_id(event->channel),
						ftdm_channel_get_ph_span_id(event->channel),
						ftdm_channel_get_ph_id(event->channel),
						ftdm_channel_get_last_error(event->channel));
			}
			break;

		case FTDM_OOB_ALARM_CLEAR:
			{
				sig.event_id = FTDM_OOB_ALARM_CLEAR;
				ftdm_clear_flag(event->channel, FTDM_CHANNEL_SUSPENDED);
				ftdm_channel_get_alarms(event->channel, &alarmbits);
				ftdm_span_send_signal(span, &sig);
			}
			break;

		default:
			break;
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : sangoma_isdn_tap_check_state()
 * 
 * Desc  : Checks the state of the span
 *
 * Ret   : void
**********************************************************************************/
static __inline__ void sangoma_isdn_tap_check_state(ftdm_span_t *span)
{
	if (ftdm_test_flag(span, FTDM_SPAN_STATE_CHANGE)) {
		uint32_t j;
		ftdm_clear_flag_locked(span, FTDM_SPAN_STATE_CHANGE);
		for(j = 1; j <= span->chan_count; j++) {
			ftdm_channel_lock(span->channels[j]);
			ftdm_channel_advance_states(span->channels[j]);
			ftdm_channel_unlock(span->channels[j]);
		}
	}
}

/**********************************************************************************
 * Fun   : sangoma_isdn_io_read()
 *  
 * Desc  : Read message as present on respective sngisdn_tap D-channel
 *
 * Ret   : res | -1
**********************************************************************************/
static int sangoma_isdn_io_read(struct sngisdn_t *isdn_msg, void *buf, int buflen)
{
	int res;
	ftdm_status_t zst;
	sngisdn_tap_t *tap = sangoma_isdn_get_userdata(isdn_msg);
	ftdm_size_t len = buflen;

	/* Read the message as present on D-Channel */
	if ((zst = ftdm_channel_read(tap->dchan, buf, &len)) != FTDM_SUCCESS) {
		if (zst == FTDM_FAIL) {
			ftdm_log(FTDM_LOG_CRIT, "Span %d D channel read fail! [%s]\n", tap->span->span_id, tap->dchan->last_error);
		} else {
			ftdm_log(FTDM_LOG_CRIT, "Span %d D channel read timeout!\n", tap->span->span_id);
		}
		return -1;
	}

	/* Getting length of the message read */
	res = (int)len;

	memset(&((unsigned char*)buf)[res],0,2);
	res += 2;

	/* Print Q921 message if DEBUG Level is set to DEBUG_Q921_RAW by the user  */
	if (tap->debug & DEBUG_Q921_RAW) {
		char hbuf[2048] = { 0 };

		print_hex_bytes(buf, len, hbuf, sizeof(hbuf));
		ftdm_log_chan(tap->dchan, FTDM_LOG_DEBUG, "READ %"FTDM_SIZE_FMT"\n%s\n", len, hbuf);
	}

	return res;
}

/**********************************************************************************
 * Fun   : sangoma_isdn_tap__get_pcall_bycrv()
 *
 * Desc  : Get sngisdn_tap call reference value if the message is respect to the
 * 	   call already in connected state
 *
 * Ret   : tap->pcalls | NULL
**********************************************************************************/
static passive_call_t *sangoma_isdn_tap_get_pcall_bycrv(sngisdn_tap_t *tap, int crv)
{
	int i;
	int tstcrv;

	ftdm_mutex_lock(tap->pcalls_lock);

	/* Go through all call present and check if the reference value matches to any one of the call the return the
	 * respective call control block else return NULL */
	for (i = 0; i < ftdm_array_len(tap->pcalls); i++) {
		tstcrv = tap->pcalls[i].callref_val;
		if (tstcrv == crv) {
			if (tap->pcalls[i].inuse) {
				ftdm_mutex_unlock(tap->pcalls_lock);
				return &tap->pcalls[i];
			}

			/* This just means the crv is being re-used in another call before this one was destroyed */
			ftdm_log(FTDM_LOG_DEBUG, "Found crv %d in slot %d of span %s with call %d but is no longer in use\n",
					crv, i, tap->span->name, tap->pcalls[i].callref_val);
		}
	}

	ftdm_log(FTDM_LOG_DEBUG, "Call reference value %d not found active in span %s\n", crv, tap->span->name);
	ftdm_mutex_unlock(tap->pcalls_lock);
	return NULL;
}

/**********************************************************************************
 * Fun   : sangoma_isdn_tap_get_pcall()
 * 
 * Desc  : It mainly gives the pcall structure w.r.t the tap on which call is
 * 	   getting established on the basis of call reference value.
 *
 * Ret   : tap->pcalls | NULL
 ***********************************************************************************/
static passive_call_t *sangoma_isdn_tap_get_pcall(sngisdn_tap_t *tap, void *callref)
{
	int i;
	int crv = 0;
	int *tmp_crv = NULL;
	passive_call_t* pcall = NULL;

	ftdm_mutex_lock(tap->pcalls_lock);

	/* If callref matches to any of the existing callref value the Enable call ref slot */
	for (i = 0; i < ftdm_array_len(tap->pcalls); i++) {
		if (callref) {
			tmp_crv = (int*)callref;
			crv = *tmp_crv;
		}

		if (crv == tap->pcalls[i].callref_val) {
			if (callref == NULL) {
				tap->pcalls[i].inuse = 1;
				ftdm_log(FTDM_LOG_DEBUG, "Enabling callref slot %d in span %s\n", i, tap->span->name);
			} else if (!tap->pcalls[i].inuse) {
				ftdm_log(FTDM_LOG_DEBUG, "Recyclying callref slot %d in span %s for callref %d/%p\n",
						i, tap->span->name, crv, callref);
				memset(&tap->pcalls[i], 0, sizeof(tap->pcalls[0]));
				//tap->pcalls[i].callref = callref;
				tap->pcalls[i].callref_val = crv;
				tap->pcalls[i].inuse = 1;
			}

			ftdm_mutex_unlock(tap->pcalls_lock);
			pcall = &tap->pcalls[i];
			return pcall;
		}
	}

	ftdm_mutex_unlock(tap->pcalls_lock);
	return NULL;
}

/**********************************************************************************
 * Fun   : sangoma_isdn_tap_put_pcall()
 * 
 * Desc  : Once DISCONNECT or RELEASE COMPLETE message is received this funtion
 * 	   release the slot on which call got established
 *
 * Ret   : void
**********************************************************************************/
static void sangoma_isdn_tap_put_pcall(sngisdn_tap_t *tap, void *callref)
{
	int i;
	int crv;
	int tstcrv;
	int *tmp_crv = callref;

	if (!callref) {
		ftdm_log(FTDM_LOG_ERROR, "Cannot put pcall for null callref in span %s\n", tap->span->name);
		return;
	}

	ftdm_mutex_lock(tap->pcalls_lock);

	/* Get the call reference value */
	crv = *tmp_crv;
	
	/* Check if the crv matches to any of the present call crv. If matches then release the slot
	 * on which respective call is active */
	for (i = 0; i < ftdm_array_len(tap->pcalls); i++) {
		if (!tap->pcalls[i].callref_val) {
			continue;
		}
		tstcrv = tap->pcalls[i].callref_val;
	       	
		if (tstcrv == crv) {
			if (tap->pcalls[i].inuse) {
				ftdm_log(FTDM_LOG_DEBUG, "Releasing slot %d in span %s used by callref %d/%d\n", i,
						tap->span->name, crv, tap->pcalls[i].callref_val);
				tap->pcalls[i].inuse = 0;
				tap->pcalls[i].proceeding = 0;
				tap->pcalls[i].callref_val = 0;
				tap->pcalls[i].fchan = NULL;
			}
		}
	}

	ftdm_mutex_unlock(tap->pcalls_lock);
}

/**********************************************************************************
 * Fun   : sangoma_isdn_tap_get_fchan()
 * 
 * Desc  : Get the channel on which call needs to be tapped
 *
 * Ret   : fchan | NULL
**********************************************************************************/
static __inline__ ftdm_channel_t *sangoma_isdn_tap_get_fchan(sngisdn_tap_t *tap, passive_call_t *pcall, int channel)
{
	ftdm_channel_t *fchan = NULL;
	int idx = 0;
	int err = 0;
	uint16_t chanpos = SNG_ISDN_CHANNEL(channel);
	ftdm_log(FTDM_LOG_CRIT, "Channel position %d and total channel configured %d\n", chanpos, tap->span->chan_count);
	if (!chanpos || (chanpos > tap->span->chan_count)) {
		ftdm_log(FTDM_LOG_CRIT, "Invalid isdn tap channel %d requested in span %s\n", channel, tap->span->name);
		return NULL;
	}

	/* Check through all the configured channels matching channel ID and return the channel structure */
	for (idx = 1; idx < 128; idx++) {
		fchan = tap->span->channels[SNG_ISDN_CHANNEL(idx)];
		if ((fchan == NULL) || (fchan->physical_chan_id == channel)) {
			break;
		}
	}

	if (!fchan) {
		err = 1;
		goto done;
	}

	ftdm_channel_lock(fchan);

	if (ftdm_test_flag(fchan, FTDM_CHANNEL_INUSE)) {
		ftdm_log(FTDM_LOG_ERROR, "Channel %d requested in span %s is already in use!\n", channel, tap->span->name);
		err = 1;
		goto done;
	}

	if (ftdm_channel_open_chan(fchan) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Could not open tap channel %d requested in span %s\n", channel, tap->span->name);
		err = 1;
		goto done;
	}

	memset(&fchan->caller_data, 0, sizeof(fchan->caller_data));

	ftdm_set_string(fchan->caller_data.cid_num.digits, pcall->callingnum.digits);
	ftdm_set_string(fchan->caller_data.cid_name, pcall->callingnum.digits);
	ftdm_set_string(fchan->caller_data.dnis.digits, pcall->callednum.digits);

done:
	if (fchan) {
		ftdm_channel_unlock(fchan);
	}

	if (err) {
		return NULL;
	}

	return fchan;
}

/**********************************************************************************
 * Fun   : handle_isdn_passive_event()
 * 
 * Desc  : Handle the events(i.e. Q931 messgages) once received and start tapping
 * 	   call depending up on the events received
 *
 * Ret   : void
**********************************************************************************/
static void handle_isdn_passive_event(sngisdn_tap_t *tap, uint8_t msgType, sng_decoder_interface_t *intfCb)
{
	passive_call_t *pcall = NULL;
	passive_call_t *peerpcall = NULL;
	ftdm_channel_t *fchan = NULL;
	ftdm_channel_t *peerfchan = NULL;
	int crv = 0;

	sngisdn_tap_t *peertap = tap->peerspan->signal_data;

	if (!intfCb)
		return;

	crv = intfCb->call_ref;

	switch (msgType) {
		case Q931_SETUP:
			/* we cannot use ftdm_channel_t because we still dont know which channel will be used 
			 * (ie, flexible channel was requested), thus, we need our own list of call references */

			ftdm_log(FTDM_LOG_DEBUG, "Call SETUP recieved on span:  %s with callref value as %d\n",
					tap->span->name, crv);
			pcall = sangoma_isdn_tap_get_pcall_bycrv(tap, crv);

			if (pcall) {
				ftdm_log(FTDM_LOG_WARNING, "There is a call with callref %d already, ignoring duplicated setup event\n", crv);
				break;
			}

			/* Try to get a recycled call (ie, get the call depending up on the crv value that the PRI stack allocated previously and then
			 * re-used for the next RING event because we did not destroy it fast enough) */
			pcall = sangoma_isdn_tap_get_pcall(tap, &crv);
			if (!pcall) {
				ftdm_log(FTDM_LOG_DEBUG, "Creating new pcall with crv = %d\n", crv);
				/* ok so the call is really not known to us, let's get a new one */
				pcall = sangoma_isdn_tap_get_pcall(tap, NULL);
				if (!pcall) {
					ftdm_log(FTDM_LOG_ERROR, "Failed to get a free passive ISDN call slot for callref %d, this is a bug!\n", crv);
					break;
				}
			}

			pcall->callref_val = crv;
			pcall->proceeding = 0;
			pcall->chanId = intfCb->chan_id;

			/* If called number is present then only fill the same */
			if (strlen(intfCb->cld_num)) {
				ftdm_log(FTDM_LOG_DEBUG,"cld_num[%s]\n", intfCb->cld_num);
				ftdm_set_string(pcall->callednum.digits, intfCb->cld_num); 
			}

			/* If calling number is present then only fill the same */
			if (strlen(intfCb->clg_num)) {
				ftdm_log(FTDM_LOG_DEBUG,"clg_num[%s]\n", intfCb->clg_num);
				ftdm_set_string(pcall->callingnum.digits, intfCb->clg_num); 
				ftdm_log(FTDM_LOG_DEBUG,"clg_num[%s]\n", pcall->callingnum.digits); 
			}

			break;

		case Q931_PROGRESS:
			ftdm_log(FTDM_LOG_DEBUG, "Call PROGRESS received on span  %s with callref value as %d\n", tap->span->name, crv);
			break;

		case Q931_PROCEEDING:
			/* at this point we should know the real b chan that will be used and can therefore proceed to notify about the call, but
			 * only if a couple of call tests are passed first */
			ftdm_log(FTDM_LOG_DEBUG, "Call PROCEED receive on Span %s with callref value as %d\n", tap->span->name, crv);

			/* CALL PROCEED will come from B end (A to B call) ,
			 * peertap is refering to A party here and tap referring to B party */

			/* CP - will come on B leg , whereas SETUP received on A leg */
			/* tap , pcall , fchan = B leg
			 * peertap, peerptap , peerfchan = A leg
			 */

			/* check that we already know about this call in the peer PRI (which was the one receiving the PRI_EVENT_RING event) */
			if (!(peerpcall = sangoma_isdn_tap_get_pcall_bycrv(peertap, crv))) {
				ftdm_log(FTDM_LOG_DEBUG,
						"Ignoring Call PROCEED event on Span %s for callref value as %d since we don't know about it\n",
						tap->span->name, crv);
				break;
			}

			if (peerpcall->proceeding) {
				ftdm_log(FTDM_LOG_DEBUG, "Duplicate Call PROCEED event received with callref value as %d\n", crv);
				ftdm_log(FTDM_LOG_DEBUG, "Ignoring Duplicate event received\n");
				break;
			}
			peerpcall->proceeding = 1;

			/* This call should not be known to this ISDN yet ... */
			if ((pcall = sangoma_isdn_tap_get_pcall_bycrv(tap, crv))) {
				ftdm_log(FTDM_LOG_ERROR,
						"Ignoring subsequent proceeding in span:%s channel:%d for callref %d\n",
						tap->span->name, pcall->chanId, crv);
				break;
			}

			/* Check if the call pointer is being recycled */
			pcall = sangoma_isdn_tap_get_pcall(tap, &crv);
			if (!pcall) {
				pcall = sangoma_isdn_tap_get_pcall(tap, NULL);
				if (!pcall) {
					ftdm_log(FTDM_LOG_ERROR, "Failed to get a free peer ISDN passive call slot for callref %d in span %s, this is a bug!\n",
							crv, tap->span->name);
					break;
				}
				pcall->callref_val = crv;
				pcall->chanId = intfCb->chan_id;
			}

			ftdm_log(FTDM_LOG_DEBUG, "Call PROCEEDING on Channel %d\n", pcall->chanId);
			/* tap = B */
			fchan = sangoma_isdn_tap_get_fchan(tap, pcall, pcall->chanId);
			if (!fchan) {
				ftdm_log(FTDM_LOG_ERROR, "Proceeding requested on odd/unavailable channel %s:1 for callref %d\n",
						tap->span->name, crv);
				break;
			}

			ftdm_log(FTDM_LOG_DEBUG, "Call PROCEEDING on Channel %d\n", peerpcall->chanId);
			/* peerfchan = A party (in a to b call) */
			peerfchan = sangoma_isdn_tap_get_fchan(peertap, peerpcall, peerpcall->chanId);
			if (!peerfchan) {
				ftdm_log(FTDM_LOG_ERROR, "Proceeding requested on odd/unavailable channel %s:1 for callref %d\n",
						peertap->span->name, crv);
				break;
			}
			pcall->fchan = fchan;
			peerpcall->fchan = peerfchan;

			ftdm_log_chan(fchan, FTDM_LOG_NOTICE, "Starting new tapped call for callref %d\n", crv);

			ftdm_channel_lock(fchan);
			fchan->call_data = peerfchan; /* fchan=B leg, pointing to peerfchan(A leg) */
			ftdm_set_state(fchan, FTDM_CHANNEL_STATE_RING);
			ftdm_channel_unlock(fchan);

			ftdm_channel_lock(peerfchan);
			peerfchan->call_data = fchan; /* peerfchan=A leg(SETUP received), pointing to fchan(B leg) */
			ftdm_set_state(peerfchan, FTDM_CHANNEL_STATE_RING);
			ftdm_channel_unlock(peerfchan);

			break;

		case Q931_CONNECT:
			ftdm_log(FTDM_LOG_DEBUG, "Call CONNECT on span:%s channel:1 is received with callref value as %d\n",
					tap->span->name, crv);

			/* CONNECT will come on B leg
			 * tap , pcall , fchan = B leg
			 * */

			if (!(pcall = sangoma_isdn_tap_get_pcall_bycrv(tap, crv))) {
				ftdm_log(FTDM_LOG_DEBUG,
						"Ignoring answer on span:%s channel:1 for callref %d since we don't know about it\n",
						tap->span->name, crv);
				break;
			}

			if (!pcall->fchan) {
				ftdm_log(FTDM_LOG_ERROR,
						"Received answer on Span:%s channel:1 for callref %d but we never got a channel\n",
						tap->span->name, crv);
				break;
			}

			/* Once Call is answered then only change channel state to UP */
			ftdm_channel_lock(pcall->fchan);
			ftdm_log_chan(pcall->fchan, FTDM_LOG_NOTICE, "Tapped call was answered in state %s\n", ftdm_channel_state2str(pcall->fchan->state));
			ftdm_set_pflag(pcall->fchan, ISDN_TAP_NETWORK_ANSWER);
			ftdm_set_state(pcall->fchan, FTDM_CHANNEL_STATE_UP);
			ftdm_channel_unlock(pcall->fchan);

			/* B leg answered, move A leg state also to ANSWER */
			peerfchan = (ftdm_channel_t*)pcall->fchan->call_data;
			if (!peerfchan) {
				ftdm_log(FTDM_LOG_ERROR,"Peer channel not found \n");
				break;
			}
			ftdm_channel_lock(peerfchan);
			ftdm_log_chan(peerfchan, FTDM_LOG_NOTICE, "Peer : Tapped call was answered in state %s\n", ftdm_channel_state2str(peerfchan->state));
			ftdm_set_pflag(peerfchan, ISDN_TAP_NETWORK_ANSWER);
			ftdm_set_state(peerfchan, FTDM_CHANNEL_STATE_UP);
			ftdm_channel_unlock(peerfchan);
			break;

		case Q931_DISCONNECT:
			ftdm_log(FTDM_LOG_DEBUG, "Call DISCONNECT received on span:%s channel:1 with callref value as %d\n",
					tap->span->name, crv);

			if (!(pcall = sangoma_isdn_tap_get_pcall_bycrv(tap, crv))) {
				ftdm_log(FTDM_LOG_DEBUG, 
						"Ignoring hangup in span:%s channel:1 for callref %d since we don't know about it\n",
						tap->span->name, crv);
				break;
			}

			if (pcall->fchan) {
				fchan = pcall->fchan;
				ftdm_channel_lock(fchan);
				if (fchan->state < FTDM_CHANNEL_STATE_TERMINATING) {
					ftdm_set_state(fchan, FTDM_CHANNEL_STATE_TERMINATING);
				}
				ftdm_channel_unlock(fchan);
			}


			/* Move peer state also to disconnect */
			peerfchan = (ftdm_channel_t*)pcall->fchan->call_data;
			if (peerfchan) {
				ftdm_channel_lock(peerfchan);
				if (peerfchan->state < FTDM_CHANNEL_STATE_TERMINATING) {
					ftdm_set_state(peerfchan, FTDM_CHANNEL_STATE_TERMINATING);
				}
				ftdm_channel_unlock(peerfchan);
			}

			sangoma_isdn_tap_put_pcall(tap, &crv);
			sangoma_isdn_tap_put_pcall(peertap, &crv);
			break;

		case Q931_RELEASE:
			ftdm_log(FTDM_LOG_DEBUG, "Call RELEASE is received on span:%s channel:1 with callref value as %d\n", 
					tap->span->name, crv);
			sangoma_isdn_tap_put_pcall(tap, &crv);
			sangoma_isdn_tap_put_pcall(peertap, &crv);
			break;

		default:
			ftdm_log(FTDM_LOG_DEBUG, "Ignoring passive event %d on span %s\n", msgType, tap->span->name);
			break;

	}
}

/**********************************************************************************
 * Fun   : sangoma_isdn_tap_check_event()
 *
 * Desc  : Get the channel on which call needs to be tapped
 *
 * Ret   : fchan | NULL
**********************************************************************************/
static void sangoma_isdn_tap_check_event(ftdm_span_t* span)
{
	ftdm_event_t *event = NULL;

	/* Poll for events, e.g HW DTMF */
	while (ftdm_span_next_event(span, &event) == FTDM_SUCCESS) {
		if (event->e_type == FTDM_EVENT_OOB) {
			ftmod_isdn_process_event(span, event);
		}
	}
}

/**********************************************************************************
 * Fun   : ftdm_sangoma_isdn_tap_run()
 * 
 * Desc  : Run the sngidsn_tap thread and always keep on monitoring on the messages
 * 	   getting received on peers as well as self sngisdn_tap D-channel.
 * 	   Depending upon which tapping needs to be starting
 *
 * Ret   : void | NULL
**********************************************************************************/
static void *ftdm_sangoma_isdn_tap_run(ftdm_thread_t *me, void *obj)
{
	ftdm_span_t *span = (ftdm_span_t *) obj;
	ftdm_span_t *peer = NULL;
	sngisdn_tap_t *tap = span->signal_data;
	sngisdn_tap_t *peer_tap = NULL;
	char 	data[1024];
	sng_decoder_interface_t intf;        	/* Interface structure between decoder and app */
	uint8_t msgType = 0;    		/* Type of the message received */
	ftdm_channel_t *ftdmchan = NULL;
	ftdm_size_t len = 0;

	/* Variables for required for polling */
	unsigned waitms = 10000;
	ftdm_iterator_t *span_chaniter = NULL;
	ftdm_iterator_t *peer_span_chaniter = NULL;
	ftdm_iterator_t *citer = NULL;
	ftdm_status_t status = FTDM_SUCCESS;
	short *span_poll_events = ftdm_malloc(sizeof(short) * span->chan_count);
	short *peer_span_poll_events = ftdm_malloc(sizeof(short) * tap->peerspan->chan_count);
	int idx = 0, res = 0;

	memset(&intf,0,sizeof(intf));

	ftdm_log(FTDM_LOG_DEBUG, "Tapping ISDN thread started on span %s\n", span->name);

	/* Initializing the d-channel of the respective span */
	span_chaniter = ftdm_span_get_chan_iterator(span, NULL);
	if (!span_chaniter) {
		ftdm_log(FTDM_LOG_CRIT, "Failed to allocate channel iterator for span %s!\n", span->name);
		goto done;
	}

	/* Initializing the d-channel of the peer span */
	peer_span_chaniter = ftdm_span_get_chan_iterator(tap->peerspan, NULL);
	if (!peer_span_chaniter) {
		ftdm_log(FTDM_LOG_CRIT, "Failed to allocate channel iterator for span %s!\n", span->name);
		goto done;
	}

	tap->span = span;

	ftdm_set_flag(span, FTDM_SPAN_IN_THREAD);

	if (ftdm_channel_open(span->span_id, tap->dchan->chan_id, &tap->dchan) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to open D-channel for span %s\n", span->name);
		goto done;
	}

	if ((tap->pri = sangoma_isdn_new_cb(tap->dchan->sockfd, PRI_NETWORK, SW_NI2, tap))) {
		sangoma_isdn_tap_set_debug(tap->pri, tap->debug);
	} else {
		ftdm_log(FTDM_LOG_CRIT, "Failed to create tapping ISDN\n");
		goto done;
	}

	/* The last span starting runs the show ...
	 * This simplifies locking and avoid races by having multiple threads for a single tapped link
	 * Since both threads really handle a single tapped link there is no benefit on multi-threading, just complications ... */
	peer = tap->peerspan;
	peer_tap = peer->signal_data;
	if (!ftdm_test_flag(tap, ISDN_TAP_MASTER)) {
		ftdm_log(FTDM_LOG_DEBUG, "Running dummy thread on span %s\n", span->name);
		while (ftdm_running() && !ftdm_test_flag(span, FTDM_SPAN_STOP_THREAD)) {
			/* We are not using this thread at all, hence just keep on sleeping */
			ftdm_sleep(100);
		}
	} else {
		ftdm_log(FTDM_LOG_DEBUG, "Master tapping thread on span %s (fd1=%d, fd2=%d)\n", span->name,
				tap->dchan->sockfd, peer_tap->dchan->sockfd);

		while (ftdm_running() && !ftdm_test_flag(span, FTDM_SPAN_STOP_THREAD)) {
			len = 1000;
			waitms = 1000;
			memset(span_poll_events, 0, sizeof(short)*span->chan_count);
			memset(peer_span_poll_events, 0, sizeof(short)*peer->chan_count);

			sangoma_isdn_tap_check_state(span);
			sangoma_isdn_tap_check_state(peer);

			/* Get poll events w.r.t each channel of span */
			for (idx = 0, citer = ftdm_span_get_chan_iterator(span, span_chaniter); citer; citer = ftdm_iterator_next(citer), idx++) {
				ftdmchan = ftdm_iterator_current(citer);

				span_poll_events[idx] |= FTDM_EVENTS;
				if (FTDM_IS_VOICE_CHANNEL(ftdmchan)) {
					if (ftdm_test_flag(ftdmchan, FTDM_CHANNEL_RX_DISABLED)) {
						span_poll_events[idx] |= FTDM_READ;
						waitms = 20;
					}
				} else {
					/* We always read d-channel */
					span_poll_events[idx] |= FTDM_READ;
				}
			}

			/* Get poll events w.r.t each channel of peer span */
			for (idx = 0, citer = ftdm_span_get_chan_iterator(peer, peer_span_chaniter); citer; citer = ftdm_iterator_next(citer), idx++) {
				ftdmchan = ftdm_iterator_current(citer);

				peer_span_poll_events[idx] |= FTDM_EVENTS;
				if (FTDM_IS_VOICE_CHANNEL(ftdmchan)) {
					if (ftdm_test_flag(ftdmchan, FTDM_CHANNEL_RX_DISABLED)) {
						peer_span_poll_events[idx] |= FTDM_READ;
						waitms = 20;
					}
				} else {
					/* We always read d-channel */
					peer_span_poll_events[idx] |= FTDM_READ;
				}
			}

			/* Initializing variable */
			res = 0;
			memset(&data, 0, sizeof(data));

			/* Polling on span and Read messages if event present on d-channel */
			status = ftdm_span_poll_event(span, waitms, span_poll_events);
			switch (status) {
				case FTDM_FAIL:
					ftdm_log(FTDM_LOG_CRIT, "Failed to poll span for IO\n");
					break;

				case FTDM_TIMEOUT:
					break;

				case FTDM_SUCCESS:
					/* Check if d-channel have any data available */
					for (citer = ftdm_span_get_chan_iterator(span, span_chaniter); citer; citer = ftdm_iterator_next(citer)) {
						len = sizeof(data);
						ftdmchan = ftdm_iterator_current(citer);
						if (FTDM_IS_DCHAN(ftdmchan)) {
							if (ftdm_test_io_flag(ftdmchan, FTDM_CHANNEL_IO_READ)) {
								/* Read messages process it iff its Q931 message */
								res = sangoma_isdn_io_read(tap->pri, data, sizeof(data));

								if (res < 5 || ((int)data[4] != 8)) {
									msgType = 0;
								} else {
									/* Need to configure the switch type using dialplan */
									msgType = (uint8_t)q931_decode(tap->switch_type, (data + 4), (res - 4), &intf);
									ftdm_log(FTDM_LOG_DEBUG, "%s Message(%d) received\n", SNG_DECODE_ISDN_EVENT(msgType), msgType);
								}

								/* If its a valid Q931 message then only handle events */
								if (msgType) {
									ftdm_log(FTDM_LOG_DEBUG, "Message Received Q931\n");
									handle_isdn_passive_event(tap, msgType, &intf);
									sangoma_isdn_tap_check_state(span);
								}
							}
						}
					}
					sangoma_isdn_tap_check_event(span);
					break;

				default:
					status = 0;
					break;
			}

			/* Polling on span and Read messages if event present on d-channel */
			status = ftdm_span_poll_event(peer, waitms, peer_span_poll_events);
			switch (status) {
				case FTDM_FAIL:
					ftdm_log(FTDM_LOG_CRIT, "Failed to poll peer span for IO\n");
					break;

				case FTDM_TIMEOUT:
					break;

				case FTDM_SUCCESS:
					/* Check if d-channel have any data available */
					for (citer = ftdm_span_get_chan_iterator(peer, peer_span_chaniter); citer; citer = ftdm_iterator_next(citer)) {
						len = sizeof(data);
						ftdmchan = ftdm_iterator_current(citer);
						if (FTDM_IS_DCHAN(ftdmchan)) {
							if (ftdm_test_io_flag(ftdmchan, FTDM_CHANNEL_IO_READ)) {
								/* Read messages process it iff its Q931 message */
								res = sangoma_isdn_io_read(peer_tap->pri, data, sizeof(data));

								if (res < 5 || ((int)data[4] != 8)) {
									msgType = 0;
								} else {
									/* Need to configure the switch type using dialplan */
									msgType = (uint8_t)q931_decode(peer_tap->switch_type, (data + 4), (res - 4), &intf);
									ftdm_log(FTDM_LOG_DEBUG, "%s Message(%d) received\n", SNG_DECODE_ISDN_EVENT(msgType), msgType);
								}

								/* If its a valid Q931 message then only handle events */
								if (msgType) {
									ftdm_log(FTDM_LOG_DEBUG, "Message Received Q931\n");
									handle_isdn_passive_event(peer_tap, msgType, &intf);
									sangoma_isdn_tap_check_state(peer);
								}
							}
						}
					}
					sangoma_isdn_tap_check_event(peer);
					break;
				default:
					status = 0;
					break;
			}

		} /* end of while */
	} /* end if */

done:
	ftdm_log(FTDM_LOG_DEBUG, "Tapping ISDN thread ended on span %s\n", span->name);

	ftdm_clear_flag(span, FTDM_SPAN_IN_THREAD);
	ftdm_clear_flag(tap, ISDN_TAP_RUNNING);
	ftdm_clear_flag(tap, ISDN_TAP_MASTER);
	ftdm_iterator_free(span_chaniter);
	ftdm_iterator_free(peer_span_chaniter);
	ftdm_safe_free(span_poll_events);
	ftdm_safe_free(peer_span_poll_events);

	return NULL;
}

/**********************************************************************************
 * Fun   : ftdm_sangoma_isdn_tap_stop()
 * 
 * Desc  : Stop configured sngisdn_tap span
 *
 * Ret   : FTDM_SUCCESS | FTDM_FAIL
**********************************************************************************/
static ftdm_status_t ftdm_sangoma_isdn_tap_stop(ftdm_span_t *span)
{
	sngisdn_tap_t *tap = span->signal_data;

	if (!ftdm_test_flag(tap, ISDN_TAP_RUNNING)) {
		return FTDM_FAIL;
	}

	ftdm_set_flag(span, FTDM_SPAN_STOP_THREAD);
	/* release local isdn message control block */
	ftdm_safe_free(tap->pri);

	while (ftdm_test_flag(span, FTDM_SPAN_IN_THREAD)) {
		ftdm_sleep(100);
	}

	ftdm_mutex_destroy(&tap->pcalls_lock);
	return FTDM_SUCCESS;
}

#if 0
/**********************************************************************************
 * Fun   : ftdm_sangoma_isdn_tap_sig_read()
 *
 * Desc  : It mainly reads the isdn tap signalling messages depending up on the
 * 	   channel ID
 *
 * Ret   : FTDM_SUCCESS | FTDM_FAIL
**********************************************************************************/
static ftdm_status_t ftdm_sangoma_isdn_tap_sig_read(ftdm_channel_t *ftdmchan, void *data, ftdm_size_t size)
{
	ftdm_status_t status;
	fio_codec_t codec_func;
	ftdm_channel_t *peerchan = ftdmchan->call_data;
	sngisdn_tap_t *tap = ftdmchan->span->signal_data;
	ftdm_size_t *chanbuf;
	ftdm_size_t *peerbuf;
	ftdm_size_t *mixedbuf;
	ftdm_size_t i = 0;
	ftdm_size_t sizeread = size;

	if (!data || !size) {
		return FTDM_FAIL;
	}

	chanbuf = ftdm_malloc(sizeof(ftdm_size_t) * size);
	peerbuf = ftdm_malloc(sizeof(ftdm_size_t) * size);
	mixedbuf = ftdm_malloc(sizeof(ftdm_size_t) * size);

	if (!FTDM_IS_VOICE_CHANNEL(ftdmchan) || !ftdmchan->call_data) {
		status =  FTDM_SUCCESS;
		goto done;
	}

	if (tap->mixaudio == ISDN_TAP_MIX_SELF) {
		status = FTDM_SUCCESS;
		goto done;
	}

	if (tap->mixaudio == ISDN_TAP_MIX_PEER) {
		/* start out by clearing the self audio to make sure we don't return audio we were
		 * not supposed to in an error condition  */
		memset(data, FTDM_SILENCE_VALUE(ftdmchan), size);
	}

	if (ftdmchan->native_codec != peerchan->native_codec) {
		ftdm_log_chan(ftdmchan, FTDM_LOG_CRIT, "Invalid peer channel with format %d, ours = %d\n", 
				peerchan->native_codec, ftdmchan->native_codec);
		status =  FTDM_FAIL;
		goto done;
	}

	memcpy(chanbuf, data, size);
	status = peerchan->fio->read(peerchan, peerbuf, &sizeread);
	if (status != FTDM_SUCCESS) {
		ftdm_log_chan_msg(ftdmchan, FTDM_LOG_ERROR, "Failed to read from peer channel!\n");
		status =  FTDM_FAIL;
		goto done;
	}
	if (sizeread != size) {
		ftdm_log_chan(ftdmchan, FTDM_LOG_ERROR, "read from peer channel only %"FTDM_SIZE_FMT" bytes!\n", sizeread);
		status =  FTDM_FAIL;
		goto done;
	}

	if (tap->mixaudio == ISDN_TAP_MIX_PEER) {
		/* only the peer audio is requested */
		memcpy(data, peerbuf, size);
		status = FTDM_SUCCESS;
		goto done;
	}

	codec_func = peerchan->native_codec == FTDM_CODEC_ULAW ? fio_ulaw2slin : peerchan->native_codec == FTDM_CODEC_ALAW ? fio_alaw2slin : NULL;
	if (codec_func) {
		sizeread = size;
		codec_func(chanbuf, sizeof(chanbuf), &sizeread);
		sizeread = size;
		codec_func(peerbuf, sizeof(peerbuf), &sizeread);
	}

	for (i = 0; i < size; i++) {
		mixedbuf[i] = ftdm_saturated_add(chanbuf[i], peerbuf[i]);
	}

	codec_func = peerchan->native_codec == FTDM_CODEC_ULAW ? fio_slin2ulaw : peerchan->native_codec == FTDM_CODEC_ALAW ? fio_slin2alaw : NULL;
	if (codec_func) {
		size = sizeof(mixedbuf);
		codec_func(mixedbuf, size, &size);
	}
	memcpy(data, mixedbuf, size);

	status = FTDM_SUCCESS;

done:
	ftdm_safe_free(chanbuf);
	ftdm_safe_free(peerbuf);
	ftdm_safe_free(mixedbuf);
	return status;
}
#endif

/**********************************************************************************
 * Fun   : ftdm_sangoma_isdn_tap_start()
 * 
 * Desc  : Start configured sngisdn_tap span
 *
 * Ret   : FTDM_FAIL | ret
**********************************************************************************/
static ftdm_status_t ftdm_sangoma_isdn_tap_start(ftdm_span_t *span)
{
	ftdm_status_t ret;
	sngisdn_tap_t *tap = span->signal_data;
	sngisdn_tap_t *peer_tap = tap->peerspan->signal_data;

	if (ftdm_test_flag(tap, ISDN_TAP_RUNNING)) {
		return FTDM_FAIL;
	}

	ftdm_mutex_create(&tap->pcalls_lock);

	ftdm_clear_flag(span, FTDM_SPAN_STOP_THREAD);
	ftdm_clear_flag(span, FTDM_SPAN_IN_THREAD);

	ftdm_set_flag(tap, ISDN_TAP_RUNNING);
	if (peer_tap && ftdm_test_flag(peer_tap, ISDN_TAP_RUNNING)) {
		/* our peer already started, we're the master */
		ftdm_set_flag(tap, ISDN_TAP_MASTER);
	}
	ret = ftdm_thread_create_detached(ftdm_sangoma_isdn_tap_run, span);

	if (ret != FTDM_SUCCESS) {
		return ret;
	}

	return ret;
}

/**************************************************************************************
 * Fun   : FIO_CONFIGURE_SPAN_SIGNALING_FUNCTION(ftdm_sangoma_isdn_tap_configure_span)
 * 
 * Desc  : Configure the isdn tap span w.r.t the configuration described as in
 * 	   freetdm.conf.xml by the user
 *
 * Ret   : FTDM_SUCCESS | FTDM_FAIL
**************************************************************************************/
static FIO_CONFIGURE_SPAN_SIGNALING_FUNCTION(ftdm_sangoma_isdn_tap_configure_span)
{
	uint32_t i;
	const char *var, *val;
	const char *debug = NULL;
	sngisdn_tap_mix_mode_t mixaudio = ISDN_TAP_MIX_BOTH;
	ftdm_channel_t *dchan = NULL;
	sngisdn_tap_t *tap = NULL;
	ftdm_span_t *peerspan = NULL;
	sngisdn_tap_iface_t iface = ISDN_TAP_IFACE_UNKNOWN;
	unsigned paramindex = 0;
	int intf = NETWORK; /* default */
	int swtch_typ = SW_ETSI; /* By default switch type will always be etsi */

	if (span->trunk_type >= FTDM_TRUNK_NONE) {
		ftdm_log(FTDM_LOG_WARNING, "Invalid trunk type '%s' defaulting to T1.\n", ftdm_trunk_type2str(span->trunk_type));
		span->trunk_type = FTDM_TRUNK_T1;
	}

	for (i = 1; i <= span->chan_count; i++) {
		if (span->channels[i]->type == FTDM_CHAN_TYPE_DQ921) {
			dchan = span->channels[i];
		}
	}

	if (!dchan) {
		ftdm_log(FTDM_LOG_ERROR, "No d-channel specified in freetdm.conf!\n");
		return FTDM_FAIL;
	}

	for (paramindex = 0; ftdm_parameters[paramindex].var; paramindex++) {
		var = ftdm_parameters[paramindex].var;
		val = ftdm_parameters[paramindex].val;
		ftdm_log(FTDM_LOG_DEBUG, "Tapping ISDN key=value, %s=%s\n", var, val);

		if (!strcasecmp(var, "debug")) {
			debug = val;
		} else if (!strcasecmp(var, "mixaudio")) {
			if (ftdm_true(val) || !strcasecmp(val, "both")) {
				ftdm_log(FTDM_LOG_DEBUG, "Setting mix audio mode to 'both' for span %s\n", span->name);
				mixaudio = ISDN_TAP_MIX_BOTH;
			} else if (!strcasecmp(val, "peer")) {
				ftdm_log(FTDM_LOG_DEBUG, "Setting mix audio mode to 'peer' for span %s\n", span->name);
				mixaudio = ISDN_TAP_MIX_PEER;
			} else {
				ftdm_log(FTDM_LOG_DEBUG, "Setting mix audio mode to 'self' for span %s\n", span->name);
				mixaudio = ISDN_TAP_MIX_SELF;
			}
		} else if (!strcasecmp(var, "interface")) {
			if (!strcasecmp(val, "cpe")) {
				iface = ISDN_TAP_IFACE_CPE;
			} else if (!strcasecmp(val, "net")) {
				iface = ISDN_TAP_IFACE_NET;
			} else {
				ftdm_log(FTDM_LOG_WARNING, "Ignoring invalid tapping interface type %s\n", val);
			}
		} else if (!strcasecmp(var, "peerspan")) {
			if (ftdm_span_find_by_name(val, &peerspan) != FTDM_SUCCESS) {
				ftdm_log(FTDM_LOG_ERROR, "Invalid tapping peer span %s\n", val);
				break;
			}
		} else if (!strcasecmp(var, "interface_type")) {
			if (!strcasecmp(val, "network")) {
				intf = NETWORK;
			} else if (!strcasecmp(val, "user")) {
				intf = USER;
			} else {
				ftdm_log(FTDM_LOG_WARNING, "Ignoring invalid tapping interface type %s\n", val);
			}
		} else if (!strcasecmp(var, "switch_type")) {
				swtch_typ = ftdm_sangoma_isdn_get_switch_type(val);
				if (swtch_typ == -1) {
					ftdm_log(FTDM_LOG_ERROR, "Invalid switch type '%s' for span '%s'\n", val, span->name);
					return FTDM_FAIL;
				}
		} else {
			ftdm_log(FTDM_LOG_ERROR,  "Unknown isdn tapping parameter [%s]", var);
		}
	}

	if (!peerspan) {
		ftdm_log(FTDM_LOG_ERROR, "No valid peerspan was specified!\n");
		return FTDM_FAIL;
	}

	tap = ftdm_calloc(1, sizeof(*tap));
	if (!tap) {
		return FTDM_FAIL;
	}

	tap->debug = parse_debug(debug);
	tap->dchan = dchan;
	tap->peerspan = peerspan;
	tap->mixaudio = mixaudio;
	tap->iface = iface;
	tap->switch_type = swtch_typ;

	span->start = ftdm_sangoma_isdn_tap_start;
	span->stop = ftdm_sangoma_isdn_tap_stop;
	span->sig_read = NULL;
	span->signal_cb = sig_cb;

	span->signal_data = tap;
	span->signal_type = FTDM_SIGTYPE_ISDN;
	span->outgoing_call = sangoma_isdn_tap_outgoing_call;

	span->get_channel_sig_status = sangoma_isdn_tap_get_channel_sig_status;
	span->get_span_sig_status = sangoma_isdn_tap_get_span_sig_status;

	span->state_map = &sngisdn_tap_state_map;
	span->state_processor = state_advance;

	/* Initialize Decoder lib */
	if (!g_decoder_lib_init) {

		sng_event.sng_log = handle_isdn_tapping_log;

		/* Initialize Decoder lib */
		if (sng_decoder_lib_init(&sng_event)) {
			ftdm_log(FTDM_LOG_ERROR, " Failed to initialize decoder lib \n");
			return FTDM_FAIL;
		}

		/* Initialize Q931 Message Decoder */
		if (q931_decoder_init(intf)) {
			ftdm_log(FTDM_LOG_ERROR, " Failed to initialize decoder lib \n");
			return FTDM_FAIL;
		}
		g_decoder_lib_init = 0x01;
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : sangoma_isdn_new_cb()
 * 
 * Desc  : Initialize the isdn tap(sngisdn_t) new control block
 *
 * Ret   : p(i.e. pri strtcture)
**********************************************************************************/
struct sngisdn_t *sangoma_isdn_new_cb(ftdm_socket_t fd, int nodetype, int switchtype, void *userdata)
{
	struct sngisdn_t *p;

	if (!(p = ftdm_calloc(1, sizeof(*p))))
		return NULL;

	p->fd = fd; 
	p->userdata = userdata;
	p->localtype = nodetype;
	p->switchtype = switchtype;
	p->cref = 1;

	return p;
}

/**********************************************************************************
 * Fun   : sangoma_isdn_get_userdata()
 * 
 * Desc  : Get user data w.r.t the pased pri control block
 *
 * Ret   : pri | NULL
**********************************************************************************/
void *sangoma_isdn_get_userdata(struct sngisdn_t *pri)
{
	return pri ? pri->userdata : NULL;
}

/**********************************************************************************
 * Fun   : handle_isdn_tapping_log()
 *
 * Desc  : prints the LOG depending up on the LOG LEVEL
 *
 * Ret   : void
**********************************************************************************/
void handle_isdn_tapping_log(uint8_t level, char *fmt,...)
{
	char	*data;
	int		ret;
	va_list	ap;

	va_start(ap, fmt);
	ret = ftdm_vasprintf(&data, fmt, ap);
	if (ret == -1) {
		goto end;
	}

	switch (level) {
	/**************************************************************************/
	case SNG_LOGLEVEL_DEBUG:
		ftdm_log(FTDM_LOG_DEBUG, "sng_decoder_lib->%s", data);
		break;
	/**************************************************************************/
	case SNG_LOGLEVEL_WARN:
		ftdm_log(FTDM_LOG_WARNING, "sng_decoder_lib->%s", data);
		break;
	/**************************************************************************/
	case SNG_LOGLEVEL_INFO:
		ftdm_log(FTDM_LOG_INFO, "sng_decoder_lib->%s", data);
		break;
	/**************************************************************************/
	case SNG_LOGLEVEL_ERROR:
		ftdm_log(FTDM_LOG_ERROR, "sng_decoder_lib->%s", data);
		break;
	/**************************************************************************/
	case SNG_LOGLEVEL_CRIT:
		/*printf("%s",data);*/
		ftdm_log(FTDM_LOG_CRIT, "sng_decoder_lib->%s", data);
		break;
	/**************************************************************************/
	default:
		ftdm_log(FTDM_LOG_INFO, "sng_decoder_lib->%s", data);
		break;
	/**************************************************************************/
	}

end:
	ftdm_safe_free(data);
	va_end(ap);
	return;
}

/**********************************************************************************
 * Fun   : pri_set_debug()
 * 
 * Desc  : Set isdn_tap debug level
 *
 * Ret   : void
**********************************************************************************/
void sangoma_isdn_tap_set_debug(struct sngisdn_t *pri, int debug)
{
	if (!pri)
		return;
	pri->debug = debug;
}

/**********************************************************************************
 * Fun   : ftdm_sangoma_isdn_get_switch_type()
 *
 * Desc  : Get the respective switch type with respect to span
 *
 * Ret   : switch_type | By default switch type remains ETSI
**********************************************************************************/
int ftdm_sangoma_isdn_get_switch_type(const char *val)
{
	int switch_type;

	if (!strcasecmp(val, "test")) {
		switch_type = SW_TST;
	} else if (!strcasecmp(val, "ccitt")) {
		switch_type = SW_CCITT;
	} else if (!strcasecmp(val, "att5eb")) {
		switch_type = SW_ATT5EB;
	} else if (!strcasecmp(val, "att5ep")) {
		switch_type = SW_ATT5EP;
	} else if (!strcasecmp(val, "att4e")) {
		switch_type = SW_ATT4E;
	} else if (!strcasecmp(val, "ntdms100b")) {
		switch_type = SW_NTDMS100B;
	} else if (!strcasecmp(val, "ntdms100p")) {
		switch_type = SW_NTDMS100P;
	} else if (!strcasecmp(val, "vn2")) {
		switch_type = SW_VN2;
	} else if (!strcasecmp(val, "vn3")) {
		switch_type = SW_VN3;
	} else if (!strcasecmp(val, "insnet")) {
		switch_type = SW_INSNET;
	} else if (!strcasecmp(val, "tr6mpc")) {
		switch_type = SW_TR6MPC;
	} else if (!strcasecmp(val, "tr6pbx")) {
		switch_type = SW_TR6PBX;
	} else if (!strcasecmp(val, "ausb")) {
		switch_type = SW_AUSB;
	} else if (!strcasecmp(val, "ausp")) {
		switch_type = SW_AUSP;
	} else if (!strcasecmp(val, "ni1")) {
		switch_type = SW_NI1;
	} else if (!strcasecmp(val, "etsi")) {
		switch_type = SW_ETSI;
	} else if (!strcasecmp(val, "bc303tmc")) {
		switch_type = SW_BC303TMC;
	} else if (!strcasecmp(val, "bc303csc")) {
		switch_type = SW_BC303CSC;
	} else if (!strcasecmp(val, "ntdms250")) {
		switch_type = SW_NTDMS250;
	} else if (!strcasecmp(val, "ni2")) {
		switch_type = SW_NI2;
	} else if (!strcasecmp(val, "qsig")) {
		switch_type = SW_QSIG;
	} else if (!strcasecmp(val, "ntmci")) {
		switch_type = SW_NTMCI;
	} else if (!strcasecmp(val, "ntni")) {
		switch_type = SW_NTNI;
	} else if (!strcasecmp(val, "bellcore")) {
		switch_type = SW_BELLCORE;
	} else if (!strcasecmp(val, "gsm")) {
		switch_type = SW_GSM;
	} else {
		switch_type = -1;
	}

	return switch_type;
}

/**********************************************************************************
 * Macro : ftdm_module
 * 
 * Desc  : FreeTDM sangoma_isdn_tap signaling and IO module definition
 *
 * Ret   : void
 **********************************************************************************/
EX_DECLARE_DATA ftdm_module_t ftdm_module = { 
	"sangoma_isdn_tap",
	ftdm_sangoma_isdn_tap_io_init,
	ftdm_sangoma_isdn_tap_unload,
	ftdm_sangoma_isdn_tap_init,
	NULL,
	NULL,
	ftdm_sangoma_isdn_tap_configure_span,
};


/* For Emacs:
 * Local Variables:
 * mode:c
 * indent-tabs-mode:t
 * tab-width:4
 * c-basic-offset:4
 * End:
 * For VIM:
 * vim:set softtabstop=4 shiftwidth=4 tabstop=4 noet:
 */
